/*
read exif meta information from jpeg files, most importantly the thumbnail.
see http://www.exif.org/
exif-2.2 has been used.
work in progress.

todo:
- search harder for the meta-info sections.  they can (probaly) reside at more addresses.
- print the meta-info nicer.  don't get too fancy though.
- check that offsets don't fall out of range of buffer...
*/

#include <u.h>
#include <libc.h>
#include <bio.h>


ulong (*get32)(char *);
uint (*get16)(char *);

uint
leget16(char *sp)
{
	uchar *p = (uchar *)sp;
	return (uint)(p[0] | p[1]<<8);
}

ulong
leget32(char *sp)
{
	return (ulong)(leget16(sp) | leget16(sp+2)<<16);
}


uint
beget16(char *sp)
{
	uchar *p = (uchar *)sp;
	return (uint)(p[0]<<8 | p[1]);
}

ulong
beget32(char *sp)
{
	return (ulong)(beget16(sp)<<16 | beget16(sp+2));
}


enum {
	Byte =		1,
	Ascii =		2,
	Short =		3,
	Long =		4,
	Rational =		5,
	Undefined =	7,
	Slong =		9,
	Srational =	10,
};


/* todo: get more names from the exif and/or tiff specs...  determine/flag which are interesting to print */
char *names[] = {
[0x010e]		"image description",
[0x010f]		"make",
[0x0110]		"model",
[0x0112]		"orientation",
[0x011a]		"x resolution",
[0x011b]		"y resolution",
[0x0128]		"resolution unit",
[0x0132]		"date time",
[0x0213]		"YCbCr positioning",
[0x8298]		"copyright",
[0x8769]		"exif ifd pointer",
};


typedef struct Ifd Ifd;
struct Ifd {
	uint tag, type;
	ulong count;
	char offset[4];
	Ifd *next;
};


/* todo: align keys... */
void
ifdprint(Ifd *ifd, char *app)
{
	char buf[16];
	char *key;
	int i;
	char *p;
	struct unit {
		int size;
		char *name;
	};
	struct unit units[] = {
	[Byte]		{1, "byte"},
	[Ascii]		{1, "ascii"},
	[Short]		{2, "short"},
	[Long]		{4, "long"},
	[Rational]		{8, "rational"},
	[Undefined]	{1, "undefined"},
	[Slong]		{4, "signed long"},
	[Srational]		{8, "signed rational"},
	};
	int size;

	if(ifd->tag < nelem(names) && names[ifd->tag])
		key = names[ifd->tag];
	else{
		snprint(buf, sizeof buf, "%4x", ifd->tag);
		key = buf;
	}

	if(ifd->type >= nelem(units))
		sysfatal("unknown type (%d)", ifd->type);

	size = units[ifd->type].size;
	if(size * ifd->count <= 4)
		p = ifd->offset;
	else
		p = app+get32(ifd->offset);

	print("%s: ", key);
	for(i = 0; i < ifd->count; i++){
		switch(ifd->type){
		case Byte:
		case Undefined:
			print("%uc", *p);
			break;
		case Ascii:
			print("%c", *p);
			break;
		case Short:
			print("%ud", get16(p));
			break;
		case Long:
			print("%uld", get32(p));
			break;
		case Rational:
			print("%uld/%uld", get32(p), get32(p+4));
			break;
		case Slong:
			print("%ld", (long)get32(p));
			break;
		case Srational:
			print("%ld/%ld", (long)get32(p), (long)get32(p+4));
			break;
		default:
			sysfatal("unknown type in ifdprint (type=%d tag=%d)", ifd->type, ifd->tag);
		}
		p += size;
	}
	print(" (%uld %s)\n", ifd->count, units[ifd->type].name);

}


Ifd *
ifdread(char *p, ulong *nextoffset)
{
	Ifd *first, *new, *cur;
	uint i, n;

	n = get16(p);
	p += 2;

	first = nil;
	for(i = 0; i < n; i++){
		new = malloc(sizeof new[0]);
		if(!new)
			sysfatal("malloc: %r");
		new->tag = get16(p);
		p += 2;
		new->type = get16(p);
		p += 2;
		new->count = get32(p);
		p += 4;
		memmove(new->offset, p, 4);
		p += 4;
		new->next = nil;

		if(!first){
			first = cur = new;
		}else{
			cur->next = new;
			cur = new;
		}
	}

	if(nextoffset)
		*nextoffset = get32(p);

	return first;
}

Ifd *
ifdfind(Ifd *i, int tag)
{
	while(i){
		if(i->tag == tag)
			return i;
		i = i->next;
	}
	return nil;
}


void
usage(void)
{
	fprint(2, "usage: exif [-t]\n");
	exits("usage");
}

void
main(int argc, char **argv)
{
	Biobuf b;
	ulong app1length;
	char buf[12];
	char exif[] = {0xff, 0xd8, 0xff, 0xe1};
	char *p;
	char *app1;
	ulong ifd0offset, ifd1offset, ifdexifoffset;
	Ifd *ifd0, *ifd1, *ifdexif, *t, *t2;
	int fd, jpegfd;
	int tflag;

	tflag = 0;
	ARGBEGIN{
	case 't':
		tflag = 1;
		break;
	default:
		usage();
	}ARGEND

	fd = 0;
	jpegfd = 1;

	if(Binit(&b, fd, OREAD) < 0)
		sysfatal("Binit: %r");

	if(Bread(&b, buf, sizeof buf) != sizeof buf)
		sysfatal("reading: %r");

	if(memcmp(buf, exif, 4) != 0)
		sysfatal("image not jpeg with exif");

	app1length = beget16(buf+5);
	app1length -= 8;

	if(memcmp(buf+6, "Exif\0\0", 6) != 0)
		sysfatal("identifier/pad does not match");

	app1 = malloc(app1length);
	if(!app1)
		sysfatal("malloc: %r");
	if(Bread(&b, app1, app1length) != app1length)
		sysfatal("reading header: %r");
	Bterm(&b);

	p = app1;

	if(memcmp(p, "MM", 2) == 0){
		get32 = beget32;
		get16 = beget16;
	}else if(memcmp(p, "II", 2) == 0){
		//assert(memcmp(p, "II", 2) == 0);
		get32 = leget32;
		get16 = leget16;
	}else{
		sysfatal("invalid endian specifcation (\"%c%c\")", p[0], p[1]);
	}
	p += 2;

	if(get16(p) != 42)
		sysfatal("missing magic number 42 (0x%ux 0x%ux)", p[0], p[1]);
	p += 2;

	ifd0offset = get32(p);
	ifd0 = ifdread(app1+ifd0offset, &ifd1offset);

	t = ifdfind(ifd0, 0x8769);
	if(!t)
		sysfatal("missing exif offset");
	ifdexifoffset = get32(t->offset);

	if(!tflag){
		print("ifd 0:\n");
		for(t = ifd0; t; t = t->next)
			ifdprint(t, app1);
		print("\n");
	}

	ifdexif = ifdread(app1+ifdexifoffset, nil);

	if(!tflag){
		print("ifd exif:\n");
		for(t = ifdexif; t; t = t->next)
			ifdprint(t, app1);
		print("\n");
	}

	ifd1 = ifdread(app1+ifd1offset, nil);

	if(!tflag){
		fprint(2, "ifd 1:\n");
		for(t = ifd1; t; t = t->next)
			ifdprint(t, app1);
		print("\n");
	}

	t = ifdfind(ifdexif, 0xa005);
	if(t){
		Ifd *ifdextra;
		if(!tflag){
			fprint(2, "ifd extra:\n");
			ifdextra = ifdread(app1+get32(t->offset), nil);
			for(t = ifdextra; t; t = t->next)
				ifdprint(t, app1);
			print("\n");
		}
	}

	if(tflag){
		t = ifdfind(ifd1, 0x0201);
		t2 = ifdfind(ifd1, 0x0202);
		if(!t || !t2)
			sysfatal("missing jpeg thumbnail\n");
		if(write(jpegfd, app1+get32(t->offset), get32(t2->offset)) != get32(t2->offset))
			sysfatal("writing jpeg: %r");
	}

	exits(0);
}
